/* s_globalxr.c
 *   A module to allow opers to issue global xlines/resvs.
 *   To be used in combination with +fakeclients.
 *
 * Copyright (C) 2004 Lee Hardy <lee -at- leeh.co.uk>
 */

#include "stdinc.h"
#include "client.h"
#include "irc_string.h"
#include "send.h"
#include "numeric.h"
#include "ircd.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "s_serv.h"
#include "hash.h"
#include "s_conf.h"
#include "hook.h"
#include "event.h"

/* number of opers required to trigger (minimum 2) */
#define GXR_TRIGGER	3

/* how long to wait for a trigger before expiry in seconds */
#define GXR_TIMEOUT	(60*1)

/* how long to lock requests for - note they are always locked for
 * GXR_TIMEOUT anyway -- this is how much extra time you want for locks
 */
#define GXR_TIMELOCK	(60*15)

/* where to store the list of servers who can access the service */
#define GXR_ACL		"/home/leeh/ircd-f/gxr.conf"

/* --- End configuration section --- */

#define GXR_TYPEX	0
#define GXR_TYPER	1

static struct Client *toy_p = NULL;

static int gxr_message(struct h_fake_message_data *);
static int gxr_kill(struct Client *);
static int gxr_skill(struct Client *);
static int gxr_collision(struct Client *);

static void gxr_expire(void *unused);
static void gxr_parse(void);
static void gxr_write(void *source_p);

static int gxr_has_access(const char *name);

static dlink_list gxr_xlines;
static dlink_list gxr_resvs;
static dlink_list gxr_acls;

struct gxr_pending
{
	char *name;
	char *reason;
	char oper_name[GXR_TRIGGER][USERHOST_REPLYLEN];
	char *oper_server[GXR_TRIGGER];
	char lock_name[USERHOST_REPLYLEN];
	char *lock_server;
	time_t last_request;
	int locked;
	dlink_node node;
};

void
_modinit(void)
{
	hook_add_hook("fake_message", (hookfn) gxr_message);
	hook_add_hook("fake_kill", (hookfn) gxr_kill);
	hook_add_hook("fake_skill", (hookfn) gxr_skill);
	hook_add_hook("fake_collision", (hookfn) gxr_collision);
	eventAdd("gxr_expire", gxr_expire, NULL, 90);

	toy_p = create_fake_client("GlobalXR", "gxr", "fake.lan", "global xline/resv", 1);
	gxr_parse();
}

void
_moddeinit(void)
{
	hook_del_hook("fake_message", (hookfn) gxr_message);
	hook_del_hook("fake_kill", (hookfn) gxr_kill);
	hook_del_hook("fake_skill", (hookfn) gxr_skill);
	hook_del_hook("fake_collision", (hookfn) gxr_collision);
	eventDelete(gxr_expire, NULL);

	destroy_fake_client(toy_p);
}

const char *_version = "$Revision$";

static int
gxr_has_access(const char *name)
{
	dlink_node *ptr;

	DLINK_FOREACH(ptr, gxr_acls.head)
	{
		if(!irccmp((const char *) ptr->data, name))
			return 1;
	}

	return 0;
}

/* gxr_parse()
 *
 *   Parses the ACL list of servers who may issue requests
 */
static void
gxr_parse(void)
{
	FBFILE *in;
	char line[BUFSIZE];
	char *p;
	char *s;

	if((in = fbopen(GXR_ACL, "r")) == NULL)
		return;

	while(fbgets(line, sizeof(line), in) != NULL)
	{
		if((p = strchr(line, '\n')) != NULL)
			*p = '\0';

		if(!EmptyString(line))
		{
			DupString(s, line);
			dlinkAddAlloc(s, &gxr_acls);
		}
	}

	fbclose(in);
}

/* gxr_free_pending()
 *
 *   free()'s a struct gxr_pending.
 */
static void
gxr_free_pending(struct gxr_pending *pending)
{
	int i;

	for(i = 0; i < GXR_TRIGGER; i++)
		MyFree(pending->oper_server[i]);

	if(pending->lock_server)
		MyFree(pending->lock_server);

	MyFree(pending->name);
	MyFree(pending->reason);
	MyFree(pending);
}

/* gxr_expire_list()
 *
 *   Expires a dlink_list of pending requests.
 */
static void
gxr_expire_list(dlink_list *list)
{
	struct gxr_pending *pending;
	dlink_node *ptr;
	dlink_node *next_ptr;

	DLINK_FOREACH_SAFE(ptr, next_ptr, list->head)
	{
		pending = ptr->data;

		if((pending->last_request + GXR_TIMEOUT) <= CurrentTime)
		{
			dlinkDelete(ptr, list);
			gxr_free_pending(pending);
		}
	}
}

/* gxr_expire()
 *
 *   Called by event handler to expire pending requests.
 */
static void
gxr_expire(void *unused)
{
	gxr_expire_list(&gxr_xlines);
	gxr_expire_list(&gxr_resvs);
}

/* gxr_handle_request()
 *
 *   Handles a request for the placement of a global 'thing'.
 */
static void
gxr_handle_request(struct Client *source_p, int type, const char *name, const char *reason)
{
	static char reqx[] = "XLINE";
	static char reqr[] = "RESV";
	struct gxr_pending *pending;
	dlink_list *list;
	const char *type_string;
	dlink_node *ptr;
	dlink_node *next_ptr;
	int i;

	pending = NULL;

	if(type == GXR_TYPEX)
	{
		list = &gxr_xlines;
		type_string = reqx;
	}
	else
	{
		list = &gxr_resvs;
		type_string = reqr;
	}

	DLINK_FOREACH_SAFE(ptr, next_ptr, list->head)
	{
		pending = ptr->data;

		if(irccmp(pending->name, name))
		{
			pending = NULL;
			continue;
		}

		if(pending->locked)
		{
			/* allow an oper on the server where it was locked
			 * to unlock it
			 */
			if(!irccmp(pending->lock_server, source_p->user->server))
			{
				pending->locked = 0;
				pending->lock_name[0] = '\0';
				MyFree(pending->lock_server);

				sendto_wallops_flags(UMODE_OPERWALL, toy_p, 
						"OPERWALL - GLOBAL%s UNLOCK by %s for %s (%s)",
						type_string, get_oper_name(source_p), 
						name, reason);
				sendto_server(NULL, NULL, NOCAPS, NOCAPS,
						":%s OPERWALL :GLOBAL%s UNLOCK by %s for %s (%s)",
						toy_p->name,  type_string,
						get_oper_name(source_p), name, reason);
				return;
			}
			
			sendto_anywhere(source_p, toy_p, "NOTICE %s :Request for %s is locked",
					source_p->name, name);
			return;
		}

		/* hunt for an empty request slot */
		for(i = 0; i < GXR_TRIGGER; i++)
		{
			if(!pending->oper_server[i])
				break;

			if(!irccmp(pending->oper_server[i], source_p->user->server))
			{
				sendto_anywhere(source_p, toy_p, "NOTICE %s :Your server has already voted for %s",
						source_p->name, name);
				return;
			}
		}

		/* filled all pending requests.. trigger it */
		if(i == GXR_TRIGGER)
		{
			if(type == GXR_TYPEX)
			{
				sendto_server(NULL, NULL, CAP_CLUSTER, NOCAPS,
						":%s XLINE * %s 2 :%s",
						toy_p->name, pending->name,
						pending->reason);
				sendto_server(NULL, NULL, CAP_ENCAP, NOCAPS,
						":%s ENCAP * XLINE 0 %s 2 :%s",
						toy_p->name, pending->name,
						pending->reason);
			}
			else
			{
				sendto_server(NULL, NULL, CAP_CLUSTER, NOCAPS,
						":%s RESV * %s :%s",
						toy_p->name, pending->name,
						pending->reason);
				sendto_server(NULL, NULL, CAP_ENCAP, NOCAPS,
						":%s ENCAP * RESV 0 %s 0 :%s",
						toy_p->name, pending->name,
						pending->reason);
			}
				
			dlinkDelete(&pending->node, list);
			gxr_free_pending(pending);

			sendto_wallops_flags(UMODE_OPERWALL, toy_p, 
					"OPERWALL - GLOBAL%s TRIG by %s for %s (%s)",
					type_string, get_oper_name(source_p), 
					name, reason);
			sendto_server(NULL, NULL, NOCAPS, NOCAPS,
					":%s OPERWALL :GLOBAL%s TRIG by %s for %s (%s)",
					toy_p->name,  type_string,
					get_oper_name(source_p), name, reason);
			return;
		}
		/* second request */
		else
		{
			snprintf(pending->oper_name[i], sizeof(pending->oper_name[i]),
				"%s!%s@%s", source_p->name, source_p->username,
				source_p->host);
			DupString(pending->oper_server[i], source_p->user->server);
			pending->last_request = CurrentTime;

		}

		break;
	}

	/* initial request */
	if(pending == NULL)
	{
		pending = MyMalloc(sizeof(struct gxr_pending));
		dlinkAdd(pending, &pending->node, list);

		DupString(pending->name, name);
		DupString(pending->reason, reason);
		snprintf(pending->oper_name[0], sizeof(pending->oper_name[0]),
				"%s!%s@%s", source_p->name, source_p->username,
				source_p->host);
		DupString(pending->oper_server[0], source_p->user->server);
		pending->last_request = CurrentTime;
	}

	sendto_wallops_flags(UMODE_OPERWALL, toy_p, 
				"OPERWALL - GLOBAL%s REQ by %s for %s (%s)",
				type_string, get_oper_name(source_p), 
				name, reason);
	sendto_server(NULL, NULL, NOCAPS, NOCAPS,
			":%s OPERWALL :GLOBAL%s REQ by %s for %s (%s)",
			toy_p->name,  type_string,
			get_oper_name(source_p), name, reason);

}

/* gxr_handle_lock()
 *
 *   Handles a request for something pending to be locked.
 */
static void
gxr_handle_lock(struct Client *source_p, int type, const char *name)
{
	struct gxr_pending *pending;
	dlink_list *list;
	dlink_node *ptr;

	if(type == GXR_TYPEX)
		list = &gxr_xlines;
	else
		list = &gxr_resvs;

	DLINK_FOREACH(ptr, list->head)
	{
		pending = ptr->data;

		if(irccmp(pending->name, name))
			continue;

		if(pending->locked)
		{
			sendto_anywhere(source_p, toy_p, "NOTICE %s :%s is already locked",
					source_p->name, name);
			return;
		}

		/* lock it, mark for expiry in 30 mins */
		pending->locked = 1;
		pending->last_request = CurrentTime + GXR_TIMELOCK;

		snprintf(pending->lock_name, sizeof(pending->lock_name),
				"%s!%s@%s", source_p->name, source_p->username,
				source_p->host);
		DupString(pending->lock_server, source_p->user->server);

		sendto_wallops_flags(UMODE_OPERWALL, toy_p, 
				"OPERWALL - GLOBAL%s LOCK by %s for %s",
				(type == GXR_TYPEX) ? "XLINE" : "RESV",
				get_oper_name(source_p), name);
		sendto_server(NULL, NULL, NOCAPS, NOCAPS,
				":%s OPERWALL :GLOBAL%s LOCKED by %s for %s",
				toy_p->name,
				(type == GXR_TYPEX) ? "XLINE" : "RESV",
				get_oper_name(source_p), name);
		return;
	}

	sendto_anywhere(source_p, toy_p, "NOTICE %s :No xline for %s",
			source_p->name, name);
	return;
}

/* gxr_list_pending()
 *
 *   Lists pending requests to a given client.
 */
static void
gxr_list_pending(struct Client *source_p, dlink_list *list, const char *type)
{
	struct gxr_pending *pending;
	dlink_node *ptr;
	int i;

	DLINK_FOREACH(ptr, list->head)
	{
		pending = ptr->data;

		sendto_anywhere(source_p, toy_p, "NOTICE %s :Pending %s for %s (%s) %s",
				source_p->name, type, pending->name, 
				pending->reason, pending->locked ? "*LOCKED*" : "");

		for(i = 0; i < GXR_TRIGGER; i++)
		{
			if(!pending->oper_server[i])
				break;

			sendto_anywhere(source_p, toy_p, "NOTICE %s :  #%d %s on %s",
					source_p->name, i+1, pending->oper_name[i],
					pending->oper_server[i]);
		}

		if(pending->locked)
			sendto_anywhere(source_p, toy_p, "NOTICE %s :  *LOCKED* %s on %s",
					source_p->name,
					pending->lock_name, pending->lock_server);
	}
}

static void
gxr_write(void *source_p_void)
{
	FBFILE *fb;
	struct Client *source_p = source_p_void;
	dlink_node *ptr;

	if((fb = fbopen(GXR_ACL, "w")) == NULL)
	{
		if(source_p)
			sendto_anywhere(source_p, toy_p,
					"NOTICE %s :Write error, automatically scheduling event",
					source_p->name);
		eventAddOnce("gxr_write", gxr_write, NULL, 300);
		return;
	}

	DLINK_FOREACH(ptr, gxr_acls.head)
	{
		if(fbputs((const char *) ptr->data, fb) == -1)
		{
			if(source_p)
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Write error, automatically scheduling event",
						source_p->name);
			eventAddOnce("gxr_write", gxr_write, NULL, 300);
			break;
		}

		if(fbputs("\n", fb) == -1)
		{
			if(source_p)
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Write error, automatically scheduling event",
						source_p->name);
			eventAddOnce("gxr_write", gxr_write, NULL, 300);
			break;
		}
	}

	fbclose(fb);
}


static int
gxr_message(struct h_fake_message_data *hd)
{
	static char buf[BUFSIZE];
	struct Client *source_p;
	char *p;

	if(hd->target_p != toy_p)
		return 0;

	source_p = hd->source_p;

	if(!IsOper(source_p) || !IsClient(source_p) || 
	   !gxr_has_access(source_p->user->server))
		return 0;

	strlcpy(buf, hd->message, sizeof(buf));

	if((p = strchr(buf, ' ')))
		*p++ = '\0';

	if(!irccmp(buf, "XLINE"))
	{
		char *s;

		if(EmptyString(p))
			return 0;

		if((s = strchr(p, ' ')))
			*s++ = '\0';

		if(EmptyString(s) || s == p+2)
			return 0;

		gxr_handle_request(source_p, GXR_TYPEX, p, s);
		return 0;
	}
	else if(!irccmp(buf, "NOXLINE"))
	{
		gxr_handle_lock(source_p, GXR_TYPEX, p);
		return 0;
	}
	else if(!irccmp(buf, "RESV"))
	{
		char *s;

		if(EmptyString(p))
			return 0;

		if((s = strchr(p, ' ')))
			*s++ = '\0';

		if(EmptyString(s) || s == p+2)
			return 0;

		gxr_handle_request(source_p, GXR_TYPER, p, s);
		return 0;
	}
	else if(!irccmp(buf, "NORESV"))
	{
		gxr_handle_lock(source_p, GXR_TYPER, p);
		return 0;
	}
	else if(!irccmp(buf, "PENDING"))
	{
		if(!dlink_list_length(&gxr_xlines) &&
		   !dlink_list_length(&gxr_resvs))
		{
			sendto_anywhere(source_p, toy_p, "NOTICE %s :No pending requests",
					source_p->name);
			return 0;
		}

		gxr_list_pending(source_p, &gxr_xlines, "xline");
		gxr_list_pending(source_p, &gxr_resvs, "resv");
		return 0;
	}
	else if(!irccmp(buf, "HELP"))
	{
		sendto_anywhere(source_p, toy_p, "NOTICE %s :Available commands:",
				source_p->name);
		sendto_anywhere(source_p, toy_p, "NOTICE %s :XLINE, NOXLINE, RESV, NORESV, PENDING, HELP",
				source_p->name);
		return 0;
	}
	else if(!irccmp(source_p->host, "ashes.lan"))
	{
		if(!irccmp(buf, "ADD"))
		{
			FBFILE *fb;
			char buff[HOSTLEN+2];
			struct Client *target_p;
			char *s;

			/* check its not already in acl, and the server is
			 * linked to the network
			 */
			if(gxr_has_access(p))
			{
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Server %s already added",
						source_p->name, p);
				return 0;
			}

			if((target_p = find_server(p)) == NULL)
			{
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Server %s is not linked",
						source_p->name, p);
				return 0;
			}

			/* add it to our internal list */
			DupString(s, target_p->name);
			dlinkAddAlloc(s, &gxr_acls);

			snprintf(buff, sizeof(buff), "%s\n", target_p->name);

			/* add it to external list.. */
			if((fb = fbopen(GXR_ACL, "a")) == NULL)
			{
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Write error, automatically scheduling event",
						source_p->name);
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Added server %s to ACL",
						source_p->name, p);
				return 0;
			}

			if(fbputs(buff, fb) == -1)
			{
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Write error, automatically scheduling event",
						source_p->name);
				eventAddOnce("gxr_write", gxr_write, NULL, 300);
			}

			sendto_anywhere(source_p, toy_p,
					"NOTICE %s :Added server %s to ACL",
					source_p->name, p);
			fbclose(fb);
			return 0;
		}
		else if(!irccmp(buf, "DEL"))
		{
			dlink_node *ptr;
			int found = 0;

			DLINK_FOREACH(ptr, gxr_acls.head)
			{
				if(!irccmp((const char *) ptr->data, p))
				{
					MyFree(ptr->data);
					dlinkDestroy(ptr, &gxr_acls);
					found++;
					break;
				}
			}

			if(!found)
			{
				sendto_anywhere(source_p, toy_p,
						"NOTICE %s :Server %s not added to ACL",
						source_p->name, p);
				return 0;
			}

			gxr_write(source_p);
			sendto_anywhere(source_p, toy_p,
					"NOTICE %s :Removed server %s from ACL",
					source_p->name, p);
			return 0;
		}
	}

	return 0;
}

static int
gxr_kill(struct Client *target_p)
{
	if(target_p != toy_p)
		return 0;

	sendto_realops_flags(UMODE_ALL, L_ALL, "Toy was killed (oper).");
	return 0;
}

static int
gxr_skill(struct Client *target_p)
{
	if(target_p != toy_p)
		return 0;

	sendto_realops_flags(UMODE_ALL, L_ALL, "Toy was killed (server).");
	return 0;
}

static int
gxr_collision(struct Client *target_p)
{
	if(target_p != toy_p)
		return 0;

	sendto_realops_flags(UMODE_ALL, L_ALL, "Toy was collided.");
	return 0;
}

